package ru.amse.pomelov.starbugs.client;

import java.awt.Graphics;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.EnumSet;
import java.util.List;

import javax.swing.JOptionPane;

import ru.amse.pomelov.starbugs.client.gui.SelectionListener;
import ru.amse.pomelov.starbugs.client.gui.frames.GameFrame;
import ru.amse.pomelov.starbugs.client.gui.tools.AbstractTool;
import ru.amse.pomelov.starbugs.remote.GameServer;
import ru.amse.pomelov.starbugs.remote.RemoteAdapter;
import ru.amse.pomelov.starbugs.remote.RemoteGameServerListener;
import ru.amse.pomelov.starbugs.server.GameField;
import ru.amse.pomelov.starbugs.turns.AbstractTurn;
import ru.amse.pomelov.starbugs.turns.EndOfTurn;
import ru.amse.pomelov.starbugs.turns.SurrenderTurn;
import ru.amse.pomelov.starbugs.units.AbstractUnit;
import ru.amse.pomelov.starbugs.units.AbstractUnitTypes;
import ru.amse.pomelov.starbugs.units.Buildings;
import ru.amse.pomelov.starbugs.units.Units;
import ru.amse.pomelov.starbugs.util.FieldPosition;
import ru.amse.pomelov.starbugs.util.UserID;
import exceptions.AlreadyInGameException;
import exceptions.ClientException;
import exceptions.GameJoinException;
import exceptions.NoFreeVacanciesException;
import exceptions.ServerConnectionException;

public class ClientDispatcher {

	private class RemoteListener extends RemoteAdapter implements Remote{
		@Override
		public void createUnits(List<AbstractUnit> unitList) throws RemoteException {
			if (myGameField != null) {
				for (AbstractUnit unit : unitList) {
					myGameField.putUnit(unit);
				}
			}
			myGameFrame.createdUnits(unitList);
		}

		@Override
		public void removeUnits(List<FieldPosition> coordsList) throws RemoteException {
			if (myGameField != null) {
				for (FieldPosition coords : coordsList) {
					myGameField.removeUnit(coords);
				}
			}
			myGameFrame.removeUnits(coordsList);
		}
		
		@Override
		public void yourTurn(int timeAvailable) throws RemoteException {
			if (isInGame) {
				myGameFrame.prepareForTurn(getAllowedUnits(), getAllowedBuildings(), timeAvailable);
			} else {
				try {
					surrender();
				} catch (ClientException e) {}
			}
		}

		@Override
		public void gameStarted() throws RemoteException {
			myGameFrame.gameStarted();
		}

		public void playerJoined() throws RemoteException {
			myGameFrame.playerJioned();
		}

		@Override
		public void endOfGame(int winner) throws RemoteException {
			if (winner == -1) {
				JOptionPane.showMessageDialog(null, "Sorry, game was stoped by the main server.");
				myGameFrame.disposeUserWaiting();
			} else if (winner == myID.playerNum) {
				JOptionPane.showMessageDialog(myGameFrame, "Congradulations!\nYou win!");
			} else {
				JOptionPane.showMessageDialog(null, "Player " + winner + " wins!");
			}
			myGameFrame.dispose();
			myMenuFrameListener.endOfGame(winner);
		}
	}
	
	
	private AbstractTool myCurrentTool;
	
	private final GameServer myGameServer;
	private final GameField myGameField;
	private UserID myID = null;
	private boolean isInGame = false;
	private RemoteGameServerListener myListener = new RemoteListener();
	private FieldPosition mySelectedCell = null;
	private SelectionListener mySelectionListener = null;
	private GameFrame myGameFrame;
	private final RemoteGameServerListener myMenuFrameListener;
	
	public ClientDispatcher(GameServer gameServer, boolean monitorOnly, RemoteGameServerListener menuFrameListener) throws RemoteException, ClientException {
		if (gameServer == null) {
			throw new IllegalArgumentException("Client dispatcher should be created with an existing game server.");
		}
		
		myGameServer = gameServer;
		UnicastRemoteObject.exportObject(myListener);

		
		myGameFrame = new GameFrame(this, monitorOnly);
		myMenuFrameListener = menuFrameListener;
		myGameField = myGameServer.getGameField();
		
		if (monitorOnly) {
			monitor();
		} else {
			join();
		}
		
	}
	
	public void join() throws ClientException {
		
		if (isInGame) {
			throw new AlreadyInGameException();
		}
		
		try {
			if (!myGameServer.hasVacancy()) {
				throw new NoFreeVacanciesException();
			}
			
			myID = myGameServer.join(myListener);
			if (myID == null) {
				throw new GameJoinException();
			}

			isInGame = true;
			
		} catch (RemoteException e) {
			throw new ServerConnectionException();
		}
	}
	
	public void monitor() throws ClientException{
		try {
			myGameServer.monitor(myListener);
		} catch (RemoteException e) {
			throw new ServerConnectionException();
		}
	}
	
	public boolean clicked(FieldPosition object) {

		AbstractUnit u = myGameField.getUnit(object);
		
		if (u == null) {
			if (myCurrentTool == null) {
				return false;
			}
		} else if ((u.getType() != AbstractUnitTypes.BUILDING)
				&& (u.getPlayer() == myID.playerNum)){
			mySelectedCell = object;
			if (mySelectionListener != null) {
				mySelectionListener.selectionGained(object);
			}
			return true;
		}
		
		if (myCurrentTool != null) {
			if ((mySelectedCell == null) && (myCurrentTool.needsObject())) {
				return false;
			}
			AbstractTurn t = myCurrentTool.generateTurn(mySelectedCell, object);
			try {
				int d = myGameServer.makeTurn(myID, t);
				if (d >= 0) {
					myGameFrame.decreaseTime(d);
				}

			} catch (RemoteException e) {
				JOptionPane.showMessageDialog(myGameFrame, "Remote exception!");
			}
		}
	

		return false;
	}
		
	public void endOfTurn() throws ClientException {
		try {
			myGameServer.makeTurn(myID, new EndOfTurn());
			myGameFrame.endOfTurn();
		} catch (RemoteException e) {
			throw new ServerConnectionException();
		}
	}
	
	public void surrender() throws ClientException {
		try {
			myGameServer.makeTurn(myID, new SurrenderTurn());
			myGameFrame.surrender();
		} catch (RemoteException e) {
			throw new ServerConnectionException();
		}
	}
	
	public void setTool(AbstractTool t) {
		if (t == null) {
			myCurrentTool = null;
			mySelectedCell = null;
			mySelectionListener.selectionLost();
			return;
		}
		myCurrentTool = t;
	}
	
	public void drawToolTip(Graphics g, int x, int y, int cellSize) {
		if ((myCurrentTool != null) && (myCurrentTool.timeNeeded() <= myGameFrame.getTimeLeft())) {
			myCurrentTool.drawToolTip(g, x, y, cellSize, myGameField.getWidth(), myGameField.getHeight());
		}
	}
	
	public GameField getGameField() {
		return myGameField;
	}

	public GameServer getGameServer() {
		return myGameServer;
	}
	
	public EnumSet<Buildings> getAllowedBuildings() {
		try {
			return myGameServer.getAllowedBuildings();
		} catch (RemoteException e) {
			return EnumSet.noneOf(Buildings.class);
		}
	}

	public EnumSet<Units> getAllowedUnits() {
		try {
			return myGameServer.getAllowedUnits();
		} catch (RemoteException e) {
			return EnumSet.noneOf(Units.class);
		}
	}
	
	public FieldPosition getSelected() {
		return mySelectedCell;
	}
	
	public void setSelectionListener(SelectionListener listener) {
		mySelectionListener = listener;
	}
	
	public int getMyNum() {
		return (myID == null) ? -1 : myID.playerNum;
	}
	
	public void forceSurrender() {
		try {
			myGameFrame.dispose();
			isInGame = false;
			myMenuFrameListener.endOfGame(0);
		} catch (RemoteException e) {
		}
	}
}
