package ru.amse.pomelov.starbugs.server;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import ru.amse.pomelov.starbugs.server.testing.Visitor;
import ru.amse.pomelov.starbugs.units.AbstractUnit;
import ru.amse.pomelov.starbugs.util.FieldPosition;

public final class GameField implements Serializable {
	
	private static final long serialVersionUID = -453878503378899748L;
	
	private final Map<FieldPosition, AbstractUnit> myGameFieldMap = new HashMap<FieldPosition, AbstractUnit>();
	private final int myWidth;
	private final int myHeight;
	
	/**
	 * Creates an empty game field with specified size.
	 * @param width
	 * @param height
	 */
	public GameField(int width, int height) {
		myWidth = width;
		myHeight = height;
	}

	/**
	 * Returns the width of the game field in cells.
	 * @return width of the game field.
	 */
	public int getWidth() {
		return myWidth;
	}
	
	/**
	 * Returns the height of the game field in cells.
	 * @return height of the game field.
	 */
	public int getHeight() {
		return myHeight;
	}

	/**
	 * Returns the unit, that palces on the game field at position (x, y).
	 * @param x 
	 * @param y
	 * @return Unit at position (x, y) or <code>null</code> if the specified cell is empty.
	 * @throws IndexOutOfBoundsException if the specified position is out of the game field.
	 */
	public AbstractUnit getUnit(FieldPosition coords) throws IndexOutOfBoundsException {
		if (!checkCoordinates(coords)) {
			throw new IndexOutOfBoundsException("GameField.getUnitAtPosition");
		}
		return myGameFieldMap.get(coords);
	}
	
	/**
	 * Puts the specified unit to the specified position. 
	 * @param x
	 * @param y
	 * @param unit unit to be placed to the game field.
	 * @return unit, that was on the specified position before operation or <code>null</code> if the
	 * cell was empty.
	 * @throws IndexOutOfBoundsException if the specified position is out of the game field.
	 */
	public AbstractUnit putUnit(AbstractUnit unit) {
		FieldPosition coords = unit.getPosition();
		if (!checkCoordinates(coords)) {
			throw new IndexOutOfBoundsException("GameField.getUnitAtPosition");
		}
		return myGameFieldMap.put(coords, unit);
	}
	
	public AbstractUnit removeUnit(FieldPosition coords) {
		return myGameFieldMap.remove(coords);
	}
	
	public boolean checkCoordinates(FieldPosition coords) {
		if (coords == null) {
			return false;
		}
		int x = coords.getX();
		int y = coords.getY();		
		if ((x < 0) || (x >= myWidth) || (y < 0) || (y >= myHeight)) {
			return false; 
		}
		return true;
	}
	
	public void accept(Visitor v) {
		v.visitGameField(this);
	}

	
	public static GameField load(InputStream in) {
		GameField result = null; 
		try {
			ObjectInputStream ois = new ObjectInputStream(in);
			int w = ois.readInt();
			int h = ois.readInt();
			result = new GameField(w, h);
			for (int i = 0; i < w; i++) {
				for (int j = 0; j < h; j++) {
					result.putUnit((AbstractUnit)ois.readObject());
				}
			}
		} catch (IOException e) {
		} catch (ClassNotFoundException e) {
		} catch (Throwable e) {
		}
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof GameField)) {
			return false;
		}
		GameField gf = (GameField)obj;
		if (gf.getHeight() != this.getHeight()) {
			return false;
		}
		if (gf.getWidth() != this.getWidth()) {
			return false;
		}
		for (int i = 0; i < getWidth(); i++) {
			for (int j = 0; j < getHeight(); j++) {
				AbstractUnit mine = getUnit(FieldPosition.createPoint(i, j));
				AbstractUnit yours = gf.getUnit(FieldPosition.createPoint(i, j));
				if ((mine == null) && (yours != null)) {
					return false;
				}
				if ((mine != null) && (yours == null)) {
					return false;
				}
				if (mine != null) {
					if (!mine.equals(yours)) {
						return false;
					}
				}
			}
		}
		return true;
	}
	
	
}
